@@ -1,18 +1,11 @@
-import six
-
-import django
 from django import forms
-from django.conf import settings
 from django.core.exceptions import ValidationError
 from django.forms.models import ModelChoiceIterator
-from django.forms.util import flatatt
 from django.forms.models import ChoiceField, ModelMultipleChoiceField
+from django.forms.utils import flatatt
+from django.utils.encoding import force_text
 from django.utils.html import escape, conditional_escape
-
-try:
-    from django.utils.encoding import force_unicode
-except ImportError:
-    from django.utils.encoding import force_text as force_unicode
+from django.utils import six
 
 from generic_plus.forms import BaseGenericFileInlineFormSet, GenericForeignFileWidget
 
@@ -29,10 +22,10 @@ class CropDusterWidget(GenericForeignFileWidget):
     template = "cropduster/custom_field.html"
 
     class Media:
-        css = {'all': (u'%scropduster/css/cropduster.css?v=6' % settings.STATIC_URL,)}
+        css = {'all': ('cropduster/css/cropduster.css',)}
         js = (
-            u'%scropduster/js/jsrender.js' % settings.STATIC_URL,
-            u'%scropduster/js/cropduster.js?v=7' % settings.STATIC_URL,
+            'cropduster/js/jsrender.js',
+            'cropduster/js/cropduster.js',
         )
 
     def get_context_data(self, name, value, attrs=None, bound_field=None):
@@ -41,8 +34,12 @@ class CropDusterWidget(GenericForeignFileWidget):
         if six.callable(sizes):
             instance = getattr(getattr(bound_field, 'form', None), 'instance', None)
             related_object = ctx['instance']
-            sizes_callable = getattr(sizes, 'im_func', sizes)
+            try:
+                sizes_callable = six.get_method_function(sizes)
+            except AttributeError:
+                sizes_callable = sizes
             sizes = sizes_callable(instance, related=related_object)
+        sizes = [s for s in sizes if not getattr(s, 'is_alias', False)]
         ctx.update({
             'sizes': json.dumps(sizes),
         })
@@ -51,8 +48,22 @@ class CropDusterWidget(GenericForeignFileWidget):
 
 class ThumbChoiceIterator(ModelChoiceIterator):
 
+    def __iter__(self):
+        if self.field.empty_label is not None:
+            yield ("", self.field.empty_label)
+        if getattr(self.field, 'cache_choices', None):
+            if self.field.choice_cache is None:
+                self.field.choice_cache = [
+                    self.choice(obj) for obj in self.queryset
+                ]
+            for choice in self.field.choice_cache:
+                yield choice
+        else:
+            for obj in self.queryset:
+                yield self.choice(obj)
+
     def choice(self, obj):
-        return (obj, self.field.label_from_instance(obj))
+        return (obj.pk, self.field.label_from_instance(obj))
 
 
 class CropDusterThumbWidget(forms.SelectMultiple):
@@ -63,33 +74,42 @@ class CropDusterThumbWidget(forms.SelectMultiple):
         super(CropDusterThumbWidget, self).__init__(*args, **kwargs)
         self.model = Thumb
 
-    def render_option(self, selected_choices, option_value, option_label):
-        attrs = {}
-        if isinstance(option_value, self.model):
-            thumb = option_value
-            option_value = thumb.pk
+    def get_option_attrs(self, value):
+        if isinstance(value, self.model):
+            thumb = value
         else:
             try:
-                thumb = self.model.objects.get(pk=option_value)
+                thumb = self.model.objects.get(pk=value)
             except (TypeError, self.model.DoesNotExist):
-                thumb = None
+                return {}
 
-        if thumb:
-            # If the thumb has no images associated with it then
-            # it has not yet been saved, and so its file path has
-            # '_tmp' appended before the extension.
-            use_tmp_file = not(thumb.image_id)
-            attrs = {
-                'data-width': thumb.width,
-                'data-height': thumb.height,
-                'data-tmp-file': json.dumps(use_tmp_file),
-            }
-        option_value = force_unicode(option_value)
+        if thumb.image_id:
+            thumb_url = thumb.image.get_image_url(size_name=thumb.name)
+        else:
+            thumb_url = None
+
+        return {
+            'data-width': thumb.width,
+            'data-height': thumb.height,
+            'data-url': thumb_url,
+            'data-tmp-file': json.dumps(not(thumb.image_id)),
+        }
+
+    def create_option(self, *args, **kwargs):
+        option = super(CropDusterThumbWidget, self).create_option(*args, **kwargs)
+        option['attrs'].update(self.get_option_attrs(option['value']))
+        option['selected'] = True
+        if isinstance(option['value'], self.model):
+            option['value'] = option['value'].pk
+        return option
+
+    def render_option(self, selected_choices, option_value, option_label):
+        attrs = self.get_option_attrs(option_value)
+        if isinstance(option_value, self.model):
+            option_value = option_value.pk
+        option_value = force_text(option_value)
         if option_value in selected_choices:
             selected_html = u' selected="selected"'
-            if not self.allow_multiple_selected:
-                # Only allow for a single selection.
-                selected_choices.remove(option_value)
         else:
             selected_html = ''
         return (
@@ -97,7 +117,7 @@ class CropDusterThumbWidget(forms.SelectMultiple):
                 'value': escape(option_value),
                 'selected': selected_html,
                 'attrs': flatatt(attrs),
-                'label': conditional_escape(force_unicode(option_label)),
+                'label': conditional_escape(force_text(option_label)),
         }
 
 
@@ -127,9 +147,43 @@ class CropDusterThumbFormField(ModelMultipleChoiceField):
     choices = property(_get_choices, ChoiceField._set_choices)
 
 
+def get_cropduster_field_on_model(model, field_identifier):
+    from cropduster.fields import CropDusterField
+
+    opts = model._meta
+    m2m_fields = [f for f in opts.get_fields() if f.many_to_many and not f.auto_created]
+    if hasattr(opts, 'private_fields'):
+        # Django 1.10+
+        private_fields = opts.private_fields
+    else:
+        # Django < 1.10
+        private_fields = opts.virtual_fields
+    m2m_related_fields = set(m2m_fields + private_fields)
+
+    field_match = lambda f: (isinstance(f, CropDusterField)
+        and f.field_identifier == field_identifier)
+
+    try:
+        return [f for f in m2m_related_fields if field_match(f)][0]
+    except IndexError:
+        return None
+
+
 class CropDusterInlineFormSet(BaseGenericFileInlineFormSet):
 
-    fields = ('image', 'thumbs', 'attribution', 'attribution_link', 'caption', 'field_identifier')
+    fields = ('image', 'thumbs', 'attribution', 'attribution_link',
+        'caption', 'alt_text', 'field_identifier')
+
+    def __init__(self, *args, **kwargs):
+        super(CropDusterInlineFormSet, self).__init__(*args, **kwargs)
+        if self.instance and not self.data:
+            cropduster_field = get_cropduster_field_on_model(self.instance.__class__, self.field_identifier)
+            if cropduster_field:
+                # An order_by() is required to prevent the queryset result cache
+                # from being removed
+                self.queryset = self.queryset.order_by('pk')
+                field_file = getattr(self.instance, cropduster_field.name)
+                self.queryset._result_cache = list(filter(None, [field_file.related_object]))
 
     def _construct_form(self, i, **kwargs):
         """
@@ -148,10 +202,7 @@ class CropDusterInlineFormSet(BaseGenericFileInlineFormSet):
 
         if form.instance and form.instance.pk:
             # Set the queryset to the current list of thumbs on the image
-            if django.VERSION < (1, 7):
-                thumbs_field.queryset = form.instance.thumbs.get_query_set()
-            else:
-                thumbs_field.queryset = form.instance.thumbs.get_queryset()
+            thumbs_field.queryset = form.instance.thumbs.get_queryset()
         else:
             # Start with an empty queryset
             thumbs_field.queryset = Thumb.objects.none()
